@@ -4,28 +4,27 @@
 
 """
 
-Functions and general utilities for rule-based pose estimation. See documentation for details
+Functions and general utilities for supervised pose estimation. See documentation for details
 
 """
 
-import cv2
-import deepof.utils
-import matplotlib.pyplot as plt
-import numpy as np
 import os
-import pandas as pd
-import regex as re
-import seaborn as sns
+import pickle
 import warnings
-from itertools import combinations
-from scipy import stats
 from typing import Any, List, NewType
 
-# Ignore warning with no downstream effect
-warnings.filterwarnings("ignore", message="All-NaN slice encountered")
+import cv2
+import numpy as np
+import pandas as pd
+import regex as re
+import sklearn.pipeline
 
-# Create custom string type
-Coordinates = NewType("Coordinates", Any)
+import deepof.utils
+
+# DEFINE CUSTOM ANNOTATED TYPES #
+project = NewType("deepof_project", Any)
+coordinates = NewType("deepof_coordinates", Any)
+table_dict = NewType("deepof_table_dict", Any)
 
 
 def close_single_contact(
@@ -51,9 +50,23 @@ def close_single_contact(
         - contact_array (np.array): True if the distance between the two specified points
         is less than tol, False otherwise"""
 
-    close_contact = (
-        np.linalg.norm(pos_dframe[left] - pos_dframe[right], axis=1) * arena_abs
-    ) / arena_rel < tol
+    close_contact = None
+
+    if isinstance(right, str):
+        close_contact = (
+            np.linalg.norm(pos_dframe[left] - pos_dframe[right], axis=1) * arena_abs
+        ) / arena_rel < tol
+
+    elif isinstance(right, list):
+        close_contact = np.any(
+            [
+                (np.linalg.norm(pos_dframe[left] - pos_dframe[r], axis=1) * arena_abs)
+                / arena_rel
+                < tol
+                for r in right
+            ],
+            axis=0,
+        )
 
     return close_contact
 
@@ -112,6 +125,28 @@ def close_double_contact(
     return double_contact
 
 
+def rotate(origin, point, ang):
+    """Auxiliar function to climb_wall and sniff_object. Rotates x,y coordinates over a pivot"""
+
+    ox, oy = origin
+    px, py = point
+
+    qx = ox + np.cos(ang) * (px - ox) - np.sin(ang) * (py - oy)
+    qy = oy + np.sin(ang) * (px - ox) + np.cos(ang) * (py - oy)
+    return qx, qy
+
+
+def outside_ellipse(x, y, e_center, e_axes, e_angle, threshold=0.0):
+    """Auxiliar function to climb_wall and sniff_object. Returns True if the passed x, y coordinates
+    are outside the ellipse denoted by e_center, e_axes and e_angle, with a certain threshold"""
+
+    x, y = rotate(e_center, (x, y), np.radians(e_angle))
+
+    term_x = (x - e_center[0]) ** 2 / (e_axes[0] + threshold) ** 2
+    term_y = (y - e_center[1]) ** 2 / (e_axes[1] + threshold) ** 2
+    return term_x + term_y > 1
+
+
 def climb_wall(
     arena_type: str,
     arena: np.array,
@@ -138,22 +173,6 @@ def climb_wall(
 
     nose = pos_dict[nose]
 
-    def rotate(origin, point, angle):
-        ox, oy = origin
-        px, py = point
-
-        qx = ox + np.cos(angle) * (px - ox) - np.sin(angle) * (py - oy)
-        qy = oy + np.sin(angle) * (px - ox) + np.cos(angle) * (py - oy)
-        return qx, qy
-
-    def outside_ellipse(x, y, e_center, e_axes, e_angle, threshold=0.0):
-
-        x, y = rotate(e_center, (x, y), np.radians(e_angle))
-
-        term_x = (x - e_center[0]) ** 2 / (e_axes[0] + threshold) ** 2
-        term_y = (y - e_center[1]) ** 2 / (e_axes[1] + threshold) ** 2
-        return term_x + term_y > 1
-
     if arena_type == "circular":
         center = np.zeros(2) if centered_data else np.array(arena[0])
         axes = arena[1]
@@ -173,50 +192,152 @@ def climb_wall(
     return climbing
 
 
+def sniff_object(
+    speed_dframe: pd.DataFrame,
+    arena_type: str,
+    arena: np.array,
+    pos_dict: pd.DataFrame,
+    tol: float,
+    tol_speed: float,
+    nose: str,
+    centered_data: bool = False,
+    s_object: str = "arena",
+    animal_id: str = "",
+):
+    """Returns True if the specified mouse is sniffing an object
+
+    Parameters:
+        - speed_dframe (pandas.DataFrame): speed of body parts over time
+        - arena_type (str): arena type; must be one of ['circular']
+        - arena (np.array): contains arena location and shape details
+        - pos_dict (table_dict): position over time for all videos in a project
+        - tol (float): minimum tolerance to report a hit
+        - nose (str): indicates the name of the body part representing the nose of
+        the selected animal
+        - arena_dims (int): indicates radius of the real arena in mm
+        - centered_data (bool): indicates whether the input data is centered
+        - object (str): indicates the object that the animal is sniffing.
+        Can be one of ['arena', 'partner']
+
+    Returns:
+        - sniffing (np.array): boolean array. True if selected animal
+        is sniffing the selected object"""
+
+    nose, nosing = pos_dict[nose], True
+
+    if animal_id != "":
+        animal_id += "_"
+
+    if s_object == "arena":
+        if arena_type == "circular":
+            center = np.zeros(2) if centered_data else np.array(arena[0])
+            axes = arena[1]
+            angle = arena[2]
+
+            nosing_min = outside_ellipse(
+                x=nose["x"],
+                y=nose["y"],
+                e_center=center,
+                e_axes=axes,
+                e_angle=-angle,
+                threshold=-tol,
+            )
+            nosing_max = outside_ellipse(
+                x=nose["x"],
+                y=nose["y"],
+                e_center=center,
+                e_axes=axes,
+                e_angle=-angle,
+                threshold=tol,
+            )
+            nosing = nosing_min & (~nosing_max)
+
+    else:
+        raise NotImplementedError
+
+    speed = speed_dframe[animal_id + "Center"] < tol_speed
+    sniffing = nosing & speed
+
+    return sniffing
+
+
 def huddle(
     pos_dframe: pd.DataFrame,
     speed_dframe: pd.DataFrame,
-    tol_forward: float,
-    tol_speed: float,
-    animal_id: str = "",
+    huddle_estimator: sklearn.pipeline.Pipeline,
 ) -> np.array:
-    """Returns true when the mouse is huddling using simple rules. (!!!) Designed to
-    work with deepof's default DLC mice models; not guaranteed to work otherwise.
+    """Returns true when the mouse is huddling a pretrained model.
 
-        Parameters:
-            - pos_dframe (pandas.DataFrame): position of body parts over time
-            - speed_dframe (pandas.DataFrame): speed of body parts over time
-            - tol_forward (float): Maximum tolerated distance between ears and
-            forward limbs
-            - tol_rear (float): Maximum tolerated average distance between spine
-            body parts
-            - tol_speed (float): Maximum tolerated speed for the center of the mouse
+    Parameters:
+        - pos_dframe (pandas.DataFrame): position of body parts over time
+        - speed_dframe (pandas.DataFrame): speed of body parts over time
+        - huddle_estimator (sklearn.pipeline.Pipeline): pre-trained model to predict feature occurrence
 
-        Returns:
-            hudd (np.array): True if the animal is huddling, False otherwise
+    Returns:
+        y_huddle (np.array): 1 if the animal is huddling, 0 otherwise
+    """
+
+    # Concatenate all relevant data frames and predict using the pre-trained estimator
+    X_huddle = pd.concat([pos_dframe, speed_dframe], axis=1).to_numpy()
+    y_huddle = huddle_estimator.predict(X_huddle)
+
+    return y_huddle
+
+
+def dig(
+    pos_dframe: pd.DataFrame,
+    speed_dframe: pd.DataFrame,
+    dig_estimator: sklearn.pipeline.Pipeline,
+):
+    """Returns true when the mouse is digging using a pretrained model.
+
+    Parameters:
+        - pos_dframe (pandas.DataFrame): position of body parts over time
+        - speed_dframe (pandas.DataFrame): speed of body parts over time
+        - dig_estimator (sklearn.pipeline.Pipeline): pre-trained model to predict feature occurrence
+
+    Returns:
+        dig (np.array): True if the animal is digging, False otherwise
+    """
+
+    # Concatenate all relevant data frames and predict using the pre-trained estimator
+    X_dig = pd.concat([pos_dframe, speed_dframe], axis=1).to_numpy()
+    y_dig = dig_estimator.predict(X_dig)
+
+    return y_dig
+
+
+def look_around(
+    speed_dframe: pd.DataFrame,
+    likelihood_dframe: pd.DataFrame,
+    tol_speed: float,
+    tol_likelihood: float,
+    animal_id: str = "",
+):
+    """Returns true when the mouse is looking around using simple rules.
+
+    Parameters:
+        - speed_dframe (pandas.DataFrame): speed of body parts over time
+        - likelihood_dframe (pandas.DataFrame): likelihood of body part tracker over time,
+        as directly obtained from DeepLabCut
+        - tol_speed (float): Maximum tolerated speed for the center of the mouse
+        - tol_likelihood (float): Maximum tolerated likelihood for the nose (if the animal
+        is digging, the nose is momentarily occluded).
+
+    Returns:
+        lookaround (np.array): True if the animal is standing still and looking around, False otherwise
     """
 
     if animal_id != "":
         animal_id += "_"
 
-    forward = (
-        np.linalg.norm(
-            pos_dframe[animal_id + "Left_bhip"] - pos_dframe[animal_id + "Left_fhip"],
-            axis=1,
-        )
-        < tol_forward
-    ) | (
-        np.linalg.norm(
-            pos_dframe[animal_id + "Right_bhip"] - pos_dframe[animal_id + "Right_fhip"],
-            axis=1,
-        )
-        < tol_forward
-    )
-
     speed = speed_dframe[animal_id + "Center"] < tol_speed
-    hudd = forward & speed
+    nose_speed = speed_dframe[animal_id + "Center"] < speed_dframe[animal_id + "Nose"]
+    nose_likelihood = likelihood_dframe[animal_id + "Nose"] > tol_likelihood
 
-    return hudd
+    lookaround = speed & nose_likelihood & nose_speed
+
+    return lookaround
 
 
 def following_path(
@@ -267,6 +388,7 @@ def following_path(
         < distance_dframe[tuple(sorted([follower + "_Nose", followed + "_Nose"]))]
     )
 
+    # noinspection PyArgumentList
     follow = np.all(
         np.array([(dist_df.min(axis=1) < tol), right_orient1, right_orient2]),
         axis=0,
@@ -275,86 +397,6 @@ def following_path(
     return follow
 
 
-def single_behaviour_analysis(
-    behaviour_name: str,
-    treatment_dict: dict,
-    behavioural_dict: dict,
-    plot: int = 0,
-    stat_tests: bool = True,
-    save: str = None,
-    ylim: float = None,
-) -> list:
-    """Given the name of the behaviour, a dictionary with the names of the groups to compare, and a dictionary
-    with the actual tags, outputs a box plot and a series of significance tests amongst the groups
-
-     Parameters:
-         - behaviour_name (str): name of the behavioural trait to analize
-         - treatment_dict (dict): dictionary containing video names as keys and experimental conditions as values
-         - behavioural_dict (dict): tagged dictionary containing video names as keys and annotations as values
-         - plot (int): Silent if 0; otherwise, indicates the dpi of the figure to plot
-         - stat_tests (bool): performs FDR corrected Mann-U non-parametric tests among all groups if True
-         - save (str): Saves the produced figure to the specified file
-         - ylim (float): y-limit for the boxplot. Ignored if plot == False
-
-     Returns:
-         - beh_dict (dict): dictionary containing experimental conditions as keys and video names as values
-         - stat_dict (dict): dictionary containing condition pairs as keys and stat results as values"""
-
-    beh_dict = {condition: [] for condition in treatment_dict.keys()}
-
-    for condition in beh_dict.keys():
-        for ind in treatment_dict[condition]:
-            beh_dict[condition] += np.sum(behavioural_dict[ind][behaviour_name]) / len(
-                behavioural_dict[ind][behaviour_name]
-            )
-
-    return_list = [beh_dict]
-
-    if plot > 0:
-
-        fig, ax = plt.subplots(dpi=plot)
-
-        sns.boxplot(
-            x=list(beh_dict.keys()),
-            y=list(beh_dict.values()),
-            orient="vertical",
-            ax=ax,
-        )
-
-        ax.set_title("{} across groups".format(behaviour_name))
-        ax.set_ylabel("Proportion of frames")
-
-        if ylim is not None:
-            ax.set_ylim(ylim)
-
-        if save is not None:  # pragma: no cover
-            plt.savefig(save)
-
-        return_list.append(fig)
-
-    if stat_tests:
-        stat_dict = {}
-        for i in combinations(treatment_dict.keys(), 2):
-            # Solves issue with automatically generated examples
-            if np.any(
-                np.array(
-                    [
-                        beh_dict[i[0]] == beh_dict[i[1]],
-                        np.var(beh_dict[i[0]]) == 0,
-                        np.var(beh_dict[i[1]]) == 0,
-                    ]
-                )
-            ):
-                stat_dict[i] = "Identical sources. Couldn't run"
-            else:
-                stat_dict[i] = stats.mannwhitneyu(
-                    beh_dict[i[0]], beh_dict[i[1]], alternative="two-sided"
-                )
-        return_list.append(stat_dict)
-
-    return return_list
-
-
 def max_behaviour(
     behaviour_dframe: pd.DataFrame, window_size: int = 10, stepped: bool = False
 ) -> np.array:
@@ -373,7 +415,7 @@ def max_behaviour(
 
     speeds = [col for col in behaviour_dframe.columns if "speed" in col.lower()]
 
-    behaviour_dframe = behaviour_dframe.drop(speeds, axis=1).astype("float")
+    behaviour_dframe = behaviour_dframe.drop(speeds, axis=1).astype(float)
     win_array = behaviour_dframe.rolling(window_size, center=True).sum()
     if stepped:
         win_array = win_array[::window_size]
@@ -401,7 +443,8 @@ def get_hparameters(hparams: dict = {}) -> dict:
         "follow_frames": 10,
         "follow_tol": 5,
         "huddle_forward": 15,
-        "huddle_speed": 1,
+        "huddle_speed": 2,
+        "nose_likelihood": 0.85,
         "fps": 24,
     }
 
@@ -438,49 +481,65 @@ def frame_corners(w, h, corners: dict = {}):
 
 
 # noinspection PyDefaultArgument,PyProtectedMember
-def rule_based_tagging(
-    tracks: List,
-    videos: List,
-    coordinates: Coordinates,
-    coords: Any,
-    dists: Any,
-    speeds: Any,
-    vid_index: int,
-    arena_type: str,
-    recog_limit: int = 1,
-    path: str = os.path.join("."),
+def supervised_tagging(
+    coord_object: coordinates,
+    raw_coords: table_dict,
+    coords: table_dict,
+    dists: table_dict,
+    angs: table_dict,
+    speeds: table_dict,
+    video: str,
+    trained_model_path: str = None,
     params: dict = {},
 ) -> pd.DataFrame:
     """Outputs a dataframe with the registered motives per frame. If specified, produces a labeled
     video displaying the information in real time
 
     Parameters:
-        - tracks (list): list containing experiment IDs as strings
-        - videos (list): list of videos to load, in the same order as tracks
-        - coordinates (deepof.preprocessing.coordinates): coordinates object containing the project information
-        - coords (deepof.preprocessing.table_dict): table_dict with already processed coordinates
-        - dists (deepof.preprocessing.table_dict): table_dict with already processed distances
-        - speeds (deepof.preprocessing.table_dict): table_dict with already processed speeds
-        - vid_index (int): index in videos of the experiment to annotate
-        - path (str): directory in which the experimental data is stored
-        - recog_limit (int): number of frames to use for arena recognition (1 by default)
-        - params (dict): dictionary to overwrite the default values of the hyperparameters of the functions
-        that the rule-based pose estimation utilizes. Values can be:
-            - speed_pause (int): size of the rolling window to use when computing speeds
-            - close_contact_tol (int): maximum distance between single bodyparts that can be used to report the trait
-            - side_contact_tol (int): maximum distance between single bodyparts that can be used to report the trait
-            - follow_frames (int): number of frames during which the following trait is tracked
-            - follow_tol (int): maximum distance between follower and followed's path during the last follow_frames,
-            in order to report a detection
-            - huddle_forward (int): maximum distance between ears and forward limbs to report a huddle detection
-            - huddle_speed (int): maximum speed to report a huddle detection
+        - coord_object (deepof.data.coordinates): coordinates object containing the project information
+        - raw_coords (deepof.data.table_dict): table_dict with raw coordinates
+        - coords (deepof.data.table_dict): table_dict with already processed (centered and aligned) coordinates
+        - dists (deepof.data.table_dict): table_dict with already processed distances
+        - angs (deepof.data.table_dict): table_dict with already processed angles
+        - speeds (deepof.data.table_dict): table_dict with already processed speeds
+        - video (str): string name of the experiment to tag
+        - trained_model_path (str): path indicating where all pretrained models are located
+        - params (dict): dictionary to overwrite the default values of the parameters of the functions
+        that the rule-based pose estimation utilizes. See documentation for details.
 
     Returns:
         - tag_df (pandas.DataFrame): table with traits as columns and frames as rows. Each
         value is a boolean indicating trait detection at a given time"""
 
+    # Load pre-trained models for ML annotated traits
+    with open(
+        os.path.join(
+            trained_model_path,
+            "deepof_supervised",
+            "deepof_supervised_huddle_estimator.pkl",
+        ),
+        "rb",
+    ) as est:
+        huddle_estimator = pickle.load(est)
+    with open(
+        os.path.join(
+            trained_model_path,
+            "deepof_supervised",
+            "deepof_supervised_dig_estimator.pkl",
+        ),
+        "rb",
+    ) as est:
+        dig_estimator = pickle.load(est)
+
+    # Extract useful information from coordinates object
+    tracks = list(coord_object._tables.keys())
+    vid_index = coord_object._videos.index(video)
+
+    arena_params = coord_object._arena_params[vid_index]
+    arena_type = coord_object._arena
+
     params = get_hparameters(params)
-    animal_ids = coordinates._animal_ids
+    animal_ids = coord_object._animal_ids
     undercond = "_" if len(animal_ids) > 1 else ""
 
     try:
@@ -488,39 +547,62 @@ def rule_based_tagging(
     except IndexError:
         vid_name = tracks[vid_index]
 
-    coords = coords[vid_name]
-    dists = dists[vid_name]
-    speeds = speeds[vid_name]
-    arena_abs = coordinates.get_arenas[1][0]
-    arena, h, w = deepof.utils.recognize_arena(
-        videos, vid_index, path, recog_limit, coordinates._arena
-    )
+    raw_coords = raw_coords[vid_name].reset_index(drop=True)
+    coords = coords[vid_name].reset_index(drop=True)
+    dists = dists[vid_name].reset_index(drop=True)
+    # angs = angs[vid_name].reset_index(drop=True)
+    speeds = speeds[vid_name].reset_index(drop=True)
+    likelihoods = coord_object.get_quality()[vid_name].reset_index(drop=True)
+    arena_abs = coord_object.get_arenas[1][0]
 
     # Dictionary with motives per frame
     tag_dict = {}
 
+    # Bulk body parts
+    main_body = [
+        "Left_ear",
+        "Right_ear",
+        "Spine_1",
+        "Center",
+        "Spine_2",
+        "Left_fhip",
+        "Right_fhip",
+        "Left_bhip",
+        "Right_bhip",
+    ]
+
     def onebyone_contact(bparts: List):
         """Returns a smooth boolean array with 1to1 contacts between two mice"""
-        nonlocal coords, animal_ids, params, arena_abs, arena
+        nonlocal raw_coords, animal_ids, params, arena_abs, arena_params
+
+        try:
+            left = animal_ids[0] + bparts[0]
+        except TypeError:
+            left = [animal_ids[0] + "_" + suffix for suffix in bparts[0]]
+
+        try:
+            right = animal_ids[1] + bparts[-1]
+        except TypeError:
+            right = [animal_ids[1] + "_" + suffix for suffix in bparts[-1]]
 
         return deepof.utils.smooth_boolean_array(
             close_single_contact(
-                coords,
-                animal_ids[0] + bparts[0],
-                animal_ids[1] + bparts[-1],
+                raw_coords,
+                (left if not isinstance(left, list) else right),
+                (right if not isinstance(left, list) else left),
                 params["close_contact_tol"],
                 arena_abs,
-                arena[1][1],
+                arena_params[1][1],
             )
         )
 
     def twobytwo_contact(rev):
         """Returns a smooth boolean array with side by side contacts between two mice"""
 
-        nonlocal coords, animal_ids, params, arena_abs, arena
+        nonlocal raw_coords, animal_ids, params, arena_abs, arena_params
         return deepof.utils.smooth_boolean_array(
             close_double_contact(
-                coords,
+                raw_coords,
                 animal_ids[0] + "_Nose",
                 animal_ids[0] + "_Tail_base",
                 animal_ids[1] + "_Nose",
@@ -528,11 +610,11 @@ def rule_based_tagging(
                 params["side_contact_tol"],
                 rev=rev,
                 arena_abs=arena_abs,
-                arena_rel=arena[1][1],
+                arena_rel=arena_params[1][1],
             )
         )
 
-    def overall_speed(speeds, _id, undercond):
+    def overall_speed(ovr_speeds, _id, ucond):
         bparts = [
             "Center",
             "Spine_1",
@@ -546,7 +628,7 @@ def rule_based_tagging(
             "Right_bhip",
             "Tail_base",
         ]
-        array = speeds[[_id + undercond + bpart for bpart in bparts]]
+        array = ovr_speeds[[_id + ucond + bpart for bpart in bparts]]
         avg_speed = np.nanmedian(array[1:], axis=1)
         return np.insert(avg_speed, 0, np.nan, axis=0)
 
@@ -564,12 +646,24 @@ def rule_based_tagging(
         tag_dict[animal_ids[1] + "_nose2tail"] = onebyone_contact(
             bparts=["_Tail_base", "_Nose"]
         )
+        tag_dict[animal_ids[0] + "_nose2body"] = onebyone_contact(
+            bparts=[
+                "_Nose",
+                main_body,
+            ]
+        )
+        tag_dict[animal_ids[1] + "_nose2body"] = onebyone_contact(
+            bparts=[
+                main_body,
+                "_Nose",
+            ]
+        )
 
         for _id in animal_ids:
             tag_dict[_id + "_following"] = deepof.utils.smooth_boolean_array(
                 following_path(
                     dists,
-                    coords,
+                    raw_coords,
                     follower=_id,
                     followed=[i for i in animal_ids if i != _id][0],
                     frames=params["follow_frames"],
@@ -581,29 +675,82 @@ def rule_based_tagging(
         tag_dict[_id + undercond + "climbing"] = deepof.utils.smooth_boolean_array(
             climb_wall(
                 arena_type,
-                arena,
-                coords,
+                arena_params,
+                raw_coords,
                 params["climb_tol"],
                 _id + undercond + "Nose",
             )
         )
-        tag_dict[_id + undercond + "speed"] = overall_speed(speeds, _id, undercond)
-        tag_dict[_id + undercond + "huddle"] = deepof.utils.smooth_boolean_array(
-            huddle(
-                coords,
+        tag_dict[_id + undercond + "sniffing"] = deepof.utils.smooth_boolean_array(
+            sniff_object(
                 speeds,
-                params["huddle_forward"],
+                arena_type,
+                arena_params,
+                raw_coords,
+                params["climb_tol"],
                 params["huddle_speed"],
+                _id + undercond + "Nose",
+                s_object="arena",
                 animal_id=_id,
             )
         )
+        tag_dict[_id + undercond + "huddle"] = huddle(
+            coords.loc[  # Filter coordinates to keep only the current animal
+                :,
+                [
+                    col
+                    for col in coords.columns
+                    if col in deepof.utils.filter_columns(coords.columns, _id)
+                ],
+            ],
+            speeds.loc[  # Filter speeds to keep only the current animal
+                :,
+                [
+                    col
+                    for col in speeds.columns
+                    if col in deepof.utils.filter_columns(speeds.columns, _id)
+                ],
+            ],
+            huddle_estimator,
+        )
+        tag_dict[_id + undercond + "dig"] = dig(
+            coords.loc[  # Filter coordinates to keep only the current animal
+                :,
+                [
+                    col
+                    for col in coords.columns
+                    if col in deepof.utils.filter_columns(coords.columns, _id)
+                ],
+            ],
+            speeds.loc[  # Filter speeds to keep only the current animal
+                :,
+                [
+                    col
+                    for col in speeds.columns
+                    if col in deepof.utils.filter_columns(speeds.columns, _id)
+                ],
+            ],
+            dig_estimator,
+        )
+        tag_dict[_id + undercond + "lookaround"] = deepof.utils.smooth_boolean_array(
+            look_around(
+                speeds,
+                likelihoods,
+                params["huddle_speed"],
+                params["nose_likelihood"],
+                animal_id=_id,
+            )
+        )
+        # NOTE: It's important that speeds remain the last columns.
+        # Preprocessing for weakly supervised autoencoders relies on this
+        tag_dict[_id + undercond + "speed"] = overall_speed(speeds, _id, undercond)
 
-    tag_df = pd.DataFrame(tag_dict)
+    tag_df = pd.DataFrame(tag_dict).fillna(0).astype(float)
 
     return tag_df
 
 
-def tag_rulebased_frames(
+def tag_annotated_frames(
     frame,
     font,
     frame_speeds,
@@ -611,29 +758,32 @@ def tag_rulebased_frames(
     corners,
     tag_dict,
     fnum,
-    dims,
     undercond,
     hparams,
     arena,
     debug,
     coords,
 ):
-    """Helper function for rule_based_video. Annotates a given frame with on-screen information
+    """Helper function for annotate_video. Annotates a given frame with on-screen information
     about the recognised patterns"""
 
-    w, h = dims
-    arena, h, w = arena
+    arena, w, h = arena
 
     def write_on_frame(text, pos, col=(255, 255, 255)):
         """Partial closure over cv2.putText to avoid code repetition"""
-        return cv2.putText(frame, text, pos, font, 1, col, 2)
+        return cv2.putText(frame, text, pos, font, 0.75, col, 2)
+
+    def conditional_flag():
+        """Returns a tag depending on a condition"""
+        if frame_speeds[animal_ids[0]] > frame_speeds[animal_ids[1]]:
+            return left_flag
+        return right_flag
 
     def conditional_pos():
         """Returns a position depending on a condition"""
         if frame_speeds[animal_ids[0]] > frame_speeds[animal_ids[1]]:
             return corners["downleft"]
-        else:
-            return corners["downright"]
+        return corners["downright"]
 
     def conditional_col(cond=None):
         """Returns a colour depending on a condition"""
@@ -641,87 +791,114 @@ def tag_rulebased_frames(
             cond = frame_speeds[animal_ids[0]] > frame_speeds[animal_ids[1]]
         if cond:
             return 150, 255, 150
-        else:
-            return 150, 150, 255
+        return 150, 150, 255
+
+    # Keep track of space usage in the output video
+    # The flags are set to False as soon as the lower
+    # corners are occupied with text
+    left_flag, right_flag = True, True
+
+    if debug:
+        # Print arena for debugging
+        cv2.ellipse(
+            img=frame,
+            center=arena[0],
+            axes=arena[1],
+            angle=arena[2],
+            startAngle=0,
+            endAngle=360,
+            color=(0, 255, 0),
+            thickness=3,
+        )
+        # Print body parts for debuging
+        for bpart in coords.columns.levels[0]:
+            if not np.isnan(coords[bpart]["x"][fnum]):
+                cv2.circle(
+                    frame,
+                    (int(coords[bpart]["x"][fnum]), int(coords[bpart]["y"][fnum])),
+                    radius=3,
+                    color=(
+                        (255, 0, 0) if bpart.startswith(animal_ids[0]) else (0, 0, 255)
+                    ),
+                    thickness=-1,
+                )
+        # Print frame number
+        write_on_frame("Frame " + str(fnum), (int(w * 0.3 / 10), int(h / 1.15)))
+
+    if len(animal_ids) > 1:
+
+        if tag_dict["nose2nose"][fnum]:
+            write_on_frame("Nose-Nose", conditional_pos())
+            if frame_speeds[animal_ids[0]] > frame_speeds[animal_ids[1]]:
+                left_flag = False
+            else:
+                right_flag = False
+
+        if tag_dict[animal_ids[0] + "_nose2body"][fnum] and left_flag:
+            write_on_frame("nose2body", corners["downleft"])
+            left_flag = False
+
+        if tag_dict[animal_ids[1] + "_nose2body"][fnum] and right_flag:
+            write_on_frame("nose2body", corners["downright"])
+            right_flag = False
+
+        if tag_dict[animal_ids[0] + "_nose2tail"][fnum] and left_flag:
+            write_on_frame("Nose-Tail", corners["downleft"])
+            left_flag = False
+
+        if tag_dict[animal_ids[1] + "_nose2tail"][fnum] and right_flag:
+            write_on_frame("Nose-Tail", corners["downright"])
+            right_flag = False
+
+        if tag_dict["sidebyside"][fnum] and left_flag and conditional_flag():
+            write_on_frame(
+                "Side-side",
+                conditional_pos(),
+            )
+            if frame_speeds[animal_ids[0]] > frame_speeds[animal_ids[1]]:
+                left_flag = False
+            else:
+                right_flag = False
+
+        if tag_dict["sidereside"][fnum] and left_flag and conditional_flag():
+            write_on_frame(
+                "Side-Rside",
+                conditional_pos(),
+            )
+            if frame_speeds[animal_ids[0]] > frame_speeds[animal_ids[1]]:
+                left_flag = False
+            else:
+                right_flag = False
 
     zipped_pos = list(
         zip(
             animal_ids,
             [corners["downleft"], corners["downright"]],
             [corners["upleft"], corners["upright"]],
+            [left_flag, right_flag],
         )
     )
 
-    if len(animal_ids) > 1:
+    for _id, down_pos, up_pos, flag in zipped_pos:
 
-        if debug:
-            # Print arena for debugging
-            cv2.ellipse(frame, arena[0], arena[1], arena[2], 0, 360, (0, 255, 0), 3)
-            # Print body parts for debuging
-            for bpart in coords.columns.levels[0]:
-                if not np.isnan(coords[bpart]["x"][fnum]):
-                    cv2.circle(
-                        frame,
-                        (int(coords[bpart]["x"][fnum]), int(coords[bpart]["y"][fnum])),
-                        radius=3,
-                        color=(
-                            (255, 0, 0)
-                            if bpart.startswith(animal_ids[0])
-                            else (0, 0, 255)
-                        ),
-                        thickness=-1,
-                    )
-            # Print frame number
-            write_on_frame("Frame " + str(fnum), corners["downleft"])
+        if flag:
 
-        if tag_dict["nose2nose"][fnum] and not tag_dict["sidebyside"][fnum]:
-            write_on_frame("Nose-Nose", conditional_pos())
-        if (
-            tag_dict[animal_ids[0] + "_nose2tail"][fnum]
-            and not tag_dict["sidereside"][fnum]
-        ):
-            write_on_frame("Nose-Tail", corners["downleft"])
-        if (
-            tag_dict[animal_ids[1] + "_nose2tail"][fnum]
-            and not tag_dict["sidereside"][fnum]
-        ):
-            write_on_frame("Nose-Tail", corners["downright"])
-        if tag_dict["sidebyside"][fnum]:
-            write_on_frame(
-                "Side-side",
-                conditional_pos(),
-            )
-        if tag_dict["sidereside"][fnum]:
-            write_on_frame(
-                "Side-Rside",
-                conditional_pos(),
-            )
-        # for _id, down_pos, up_pos in zipped_pos:
-        #     if (
-        #         tag_dict[_id + "_following"][fnum]
-        #         and not tag_dict[_id + "_climbing"][fnum]
-        #     ):
-        #         write_on_frame(
-        #             "*f",
-        #             (int(w * 0.3 / 10), int(h / 10)),
-        #             conditional_col(),
-        #         )
-
-    for _id, down_pos, up_pos in zipped_pos:
-
-        if tag_dict[_id + undercond + "climbing"][fnum]:
-            write_on_frame("Climbing", down_pos)
-        if (
-            tag_dict[_id + undercond + "huddle"][fnum]
-            and not tag_dict[_id + undercond + "climbing"][fnum]
-        ):
-            write_on_frame("huddle", down_pos)
+            if tag_dict[_id + undercond + "climbing"][fnum]:
+                write_on_frame("climbing", down_pos)
+            elif tag_dict[_id + undercond + "huddle"][fnum]:
+                write_on_frame("huddling", down_pos)
+            elif tag_dict[_id + undercond + "sniffing"][fnum]:
+                write_on_frame("sniffing", down_pos)
+            elif tag_dict[_id + undercond + "dig"][fnum]:
+                write_on_frame("digging", down_pos)
+            # elif tag_dict[_id + undercond + "lookaround"][fnum]:
+            #     write_on_frame("lookaround", down_pos)
 
         # Define the condition controlling the colour of the speed display
         if len(animal_ids) > 1:
             colcond = frame_speeds[_id] == max(list(frame_speeds.values()))
         else:
-            colcond = hparams["huddle_speed"] > frame_speeds
+            colcond = hparams["huddle_speed"] < frame_speeds
 
         write_on_frame(
             str(
@@ -736,33 +913,24 @@ def tag_rulebased_frames(
 
 
 # noinspection PyProtectedMember,PyDefaultArgument
-def rule_based_video(
-    coordinates: Coordinates,
-    tracks: List,
-    videos: List,
-    vid_index: int,
+def annotate_video(
+    coordinates: coordinates,
     tag_dict: pd.DataFrame,
+    vid_index: int,
     frame_limit: int = np.inf,
-    recog_limit: int = 1,
-    path: str = os.path.join("."),
-    params: dict = {},
     debug: bool = False,
+    params: dict = {},
 ) -> True:
-    """Renders a version of the input video with all rule-based taggings in place.
+    """Renders a version of the input video with all supervised taggings in place.
 
     Parameters:
-        - tracks (list): list containing experiment IDs as strings
-        - videos (list): list of videos to load, in the same order as tracks
         - coordinates (deepof.preprocessing.coordinates): coordinates object containing the project information
         - debug (bool): if True, several debugging attributes (such as used body parts and arena) are plotted in
         the output video
-        - vid_index (int): index in videos of the experiment to annotate
-        - fps (float): frames per second of the analysed video. Same as input by default
-        - path (str): directory in which the experimental data is stored
+        - vid_index: for internal usage only; index of the video to tag in coordinates._videos
         - frame_limit (float): limit the number of frames to output. Generates all annotated frames by default
-        - recog_limit (int): number of frames to use for arena recognition (1 by default)
         - params (dict): dictionary to overwrite the default values of the hyperparameters of the functions
-        that the rule-based pose estimation utilizes. Values can be:
+        that the supervised pose estimation utilizes. Values can be:
             - speed_pause (int): size of the rolling window to use when computing speeds
             - close_contact_tol (int): maximum distance between single bodyparts that can be used to report the trait
             - side_contact_tol (int): maximum distance between single bodyparts that can be used to report the trait
@@ -771,13 +939,19 @@ def rule_based_video(
             in order to report a detection
             - huddle_forward (int): maximum distance between ears and forward limbs to report a huddle detection
             - huddle_speed (int): maximum speed to report a huddle detection
+            - fps (float): frames per second of the analysed video. Same as input by default
+
 
     Returns:
         True
 
     """
 
-    # DATA OBTENTION AND PREPARATION
+    # Extract useful information from coordinates object
+    tracks = list(coordinates._tables.keys())
+    videos = coordinates._videos
+    path = os.path.join(coordinates._path, "Videos")
+
     params = get_hparameters(params)
     animal_ids = coordinates._animal_ids
     undercond = "_" if len(animal_ids) > 1 else ""
@@ -787,9 +961,8 @@ def rule_based_video(
     except IndexError:
         vid_name = tracks[vid_index]
 
-    arena, h, w = deepof.utils.recognize_arena(
-        videos, vid_index, path, recog_limit, coordinates._arena
-    )
+    arena_params = coordinates._arena_params[vid_index]
+    h, w = coordinates._video_resolution[vid_index]
     corners = frame_corners(h, w)
 
     cap = cv2.VideoCapture(os.path.join(path, videos[vid_index]))
@@ -809,7 +982,7 @@ def rule_based_video(
             print("Can't receive frame (stream end?). Exiting ...")
             break
 
-        font = cv2.FONT_HERSHEY_COMPLEX_SMALL
+        font = cv2.FONT_HERSHEY_DUPLEX
 
         # Capture speeds
         try:
@@ -824,7 +997,7 @@ def rule_based_video(
                 frame_speeds = tag_dict["speed"][fnum]
 
         # Display all annotations in the output video
-        tag_rulebased_frames(
+        tag_annotated_frames(
             frame,
             font,
             frame_speeds,
@@ -832,10 +1005,9 @@ def rule_based_video(
             corners,
             tag_dict,
             fnum,
-            (w, h),
             undercond,
             params,
-            (arena, h, w),
+            (arena_params, h, w),
             debug,
             coordinates.get_coords(center=False)[vid_name],
         )
@@ -861,5 +1033,10 @@ def rule_based_video(
     return True
 
 
+if __name__ == "__main__":
+    # Ignore warnings with no downstream effect
+    warnings.filterwarnings("ignore", message="All-NaN slice encountered")
+
 # TODO:
-#    - Is border sniffing anything you might consider interesting?
+#    - Scale CD1 features to match those of black6
+#    - Use scaled features (not standard, but scaled using some notion of animal size)
